/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.regex.Pattern;

import com.unitesk.requality.core.TreeNodeChangeEvent.TreeNodeChangeEventType;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Attribute.ParameterAvailability;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.domproxy.LinkResolver;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.RestrictedFolder;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class TreeDB {

    public static final String ORDER_ATTRIBUTE = "_order";
    public static final String INDEX_ATTRIBUTE = "_index";
    public static final String TEST_GENERATORS_FILTERS = "_testGenFilters";
    public static final String DEFAULT_REFERENCE_NAME = "_defRefName";
    public static final String ACTIVE_TEMPLATES_ATTRIBUTE = "_active_templates";
    public static final String EMPTY_TEMPLATE = "&empty";
    private static final String PREFERENCE_FNMODE = "_com.unitesk.requality.full_name_mode";
    private List<ITreeChangeListener> listeners = new CopyOnWriteArrayList<ITreeChangeListener>();
    private List<ITreeChangeListener> guilisteners = new CopyOnWriteArrayList<ITreeChangeListener>();
    private TreeNode root;
    private Map<UUID, TreeNode> cache = new ConcurrentHashMap<UUID, TreeNode>();
    private List<String> testgenfilters;
    private ITreeStorage storage;
    private IUUIDGenerator uuidGenerator;
    private TreeLogic tlogic;
    private HashMap<String, Set<UUID>> namecache = new HashMap<String, Set<UUID>>();
    private HashMap<UUID, String> backnamecache = new HashMap<UUID, String>(); 
    private Map<String, NodeType> defaultTypes = new HashMap<String, NodeType>();
    
    private boolean isVirtual;
    
    private ConcurrentLinkedQueue<TreeNodeChangeEvent> guichanges = new ConcurrentLinkedQueue<TreeNodeChangeEvent>();
    private int activeChangesRowsCount = 0;
    
    private Map<String,String> activeTemplates = null;

    public boolean getFullNameMode(){
    	if(root == null)
    		return false;
    	Attribute a = root.getAttribute(PREFERENCE_FNMODE);
    	return a!=null && a.getValue().toString().toLowerCase().equals("true");
    }
    public void setFullNameMode(boolean useUVName){
    	if(root == null)
    		return;
    	root.putAttribute(new Attribute(root,AttributeType.BOOL,PREFERENCE_FNMODE,useUVName));
    	root.saveAttributes();
    }
    public void loadActiveTemplates(){
    	activeTemplates = new HashMap<String, String>();
    	if(getTemplatesFolder()==null)
    		return;
    	Attribute a = getTemplatesFolder().getAttribute(ACTIVE_TEMPLATES_ATTRIBUTE);
    	if(a!=null){
    		List<String> info = (List<String>) a.getValue();
    		for(String s : info){
    			String type = s.substring(0,s.indexOf("|"));
    			String id = s.substring(s.indexOf("|")+1);
    			if(id.isEmpty())
    				activeTemplates.put(type, null);
    			else{
    				TreeNode node = getTemplate(id);
    				if(node!=null)
    					activeTemplates.put(type, id);
    				else
    					activeTemplates.put(type, null);
    			}
    		}
    	}
    }
    public Map<String,String> getActiveTemplates(){
    	if(activeTemplates==null)
    		loadActiveTemplates();
    	return activeTemplates;
    }
    public void saveActiveTemplates(){
    	List<Object> infos = new ArrayList<Object>();
    	for(String s:getActiveTemplates().keySet())
    		{
    			String name = getActiveTemplates().get(s);
    			if(name!=null){
    				TreeNode tmpl = getTemplate(name);
    				if(tmpl!=null)
    					infos.add(s+"|"+tmpl.getId());
    				else
    					infos.add(s+"|");
    			}else
        			infos.add(s+"|");
    		}
    	getOrCreateTemplatesFolder().putAttribute(new Attribute(getTemplatesFolder(), 
    			AttributeType.LIST,
    			ACTIVE_TEMPLATES_ATTRIBUTE, 
    			new ListValue(AttributeType.STRING,infos)));
    	getOrCreateTemplatesFolder().saveAttributes();
    }
    private TreeNode getTemplatesFolder(){
    	return root.findChild("Templates");
    }
    public TreeNode getOrCreateTemplatesFolder(){
    	TreeNode node = getTemplatesFolder();
    	if(node == null)
    		try{
    			node = root.createChild("Templates",RestrictedFolder.TYPE_NAME,null);
    			node.addChangeListener(new INodeChangeListener() {
					@Override
					public void moved(TreeNode source, TreeNode oldParent, TreeNode newParent) {
					}
					
					@Override
					public void deleted(TreeNode source) {
					}
					
					@Override
					public void attributeChange(TreeNode source, String what) {
						if(what.equals(ACTIVE_TEMPLATES_ATTRIBUTE)){
							loadActiveTemplates();
						}
					}
				});
    			return node;
    		}
    		catch (Exception e) {
    			e.printStackTrace();
			}	
    	return node;
    }


	public TreeNode createTemplate(String type) {
		TreeNode fol = getOrCreateTemplatesFolder();
		String id = RequalityCLI.getNextId(fol, type, "Template_");
		return fol.createChild(id,type,null);
	}
    public boolean extractTemplate(TreeNode node,String name){
    	TreeNode fold = getOrCreateTemplatesFolder();
    	if(getTemplate(name)!=null)
    		return false;
    	Map<String,Attribute> map = new HashMap<String,Attribute>();
    	UUID newuid = UUID.randomUUID();
    	for(Attribute a:node.getAttributes().values())
    		if(a.getOrigin().equals(node.getUUId()))
    			map.put(a.getKey(), a.copy(newuid));
    		else
    			map.put(a.getKey(), a);
    	int num = 0;
    	String retname = name;
    	while(fold.findChild(retname)!=null){
    		retname = name+"_"+num;
    		num++;
    	}
    	NodeDesc nd = new NodeDesc(newuid, fold.getUUId(), retname, map, new ArrayList<UUID>());
    	nd.createNode(this);
    	return true;
    }

    public List<TreeNode> getTemplates(){
    	List<TreeNode> templates = new ArrayList<TreeNode>();
    	TreeNode tfolder = getTemplatesFolder();
    	if(tfolder == null)
    		return Collections.EMPTY_LIST;
    	for(TreeNode templ:tfolder)
    		templates.add(templ);
    	return templates;
    }
    
    public TreeNode getTemplate(String name){
    	if(name == null)
   		 	return null;
    	if(getTemplatesFolder()==null)
    		return null;
    	return getTemplatesFolder().findChild(name);
    }
    
    public void setActiveTemplate(String type, String templ){
    	getOrCreateTemplatesFolder();
    	getActiveTemplates().put(type,templ);
    	saveActiveTemplates();
    }
    
    public TreeNode getActiveTemplate(String type){
    	if(!getActiveTemplates().containsKey(type))
    		return null;
    	return getTemplate(getActiveTemplates().get(type));
    }
    
    
    private ConcurrentLinkedQueue<TreeNodeChangeEvent> getGuiChanges(){
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.getGuiChanges();
    	}
    	return guichanges;
    }
    {
        NodeType dbroot = new NodeType("DBRoot", TreeDB.this);
        dbroot.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("DBRoot", dbroot);
        NodeType nodetype = new NodeType("NodeType", TreeDB.this);
        nodetype.setClass(com.unitesk.requality.nodetypes.NodeType.class);
        defaultTypes.put("NodeType", nodetype);

        NodeType vnodetype = new NodeType("VirtualNode", TreeDB.this);
        vnodetype.setClass(com.unitesk.requality.nodetypes.VirtualNode.class);
        defaultTypes.put("VirtualNode", vnodetype);
        
        NodeType node = new NodeType("Node", TreeDB.this);
        node.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("Node", node);

    };

    public List<String> getTestGeneratorsFilters(){
    	if(testgenfilters == null)
    		testgenfilters = (List<String>) getRootNode().getListAttributeValue(TEST_GENERATORS_FILTERS,AttributeType.STRING);
    	return testgenfilters;
    }
    /**
     * Returns Type of element with given uid if it exists.
     * @param uid uuid of element
     * @return type of element or null
     */
    public String getTreeNodeType(UUID uid){
    	if(isLoaded(uid))
    		return getNode(uid).getType();
    	return storage.getNodeType(uid);
    }
    /**
     * Combine all future events in a row to be handled concurrently. 
     */
    public void startChangeRow(){
    	activeChangesRowsCount++;
    }
    /**
     * Returns 'virtuality' status of catalog. Catalog is virtual if 
     * it is linked to virtual storage.
     * @return true if catalog is virtual
     */
    public boolean isVirtual(){
    	return isVirtual;
    } 
    /**
     * Sets 'virtuality' status of catalog. Catalog is virtual if 
     * it is linked to virtual storage. This method is public
     * only to be available for inner mechanics. Should not be
     * called by side plugins.
     * @return true if catalog is virtual
     */
    public void setVirtualState(boolean value){
    	isVirtual = value;
    }
    /**
     * Stop event row started with startChangeRow() method
     */
    public void stopChangeRow(){
    	if(activeChangesRowsCount > 0)
    		activeChangesRowsCount--;
    	if(activeChangesRowsCount == 0){
    		HashMap<UUID,Integer>lastchange = new HashMap<UUID, Integer>();//,
    		//					 lastmove = new HashMap<UUID, Integer>();
    		HashSet<UUID> deleted = new HashSet<UUID>(),
    					  created = new HashSet<UUID>();
    		HashMap<UUID,Set<String>> changeParams = new HashMap<UUID, Set<String>>();
    		int i = 0; 
    		for(TreeNodeChangeEvent change:getGuiChanges()){
    			if(change.getType() == TreeNodeChangeEventType.DELETE){
    				NodeDesc nd = change.getNodeDescTarget();
    				if(nd!=null){
    					deleted.add(nd.getUUId());
    				}else
    					deleted.add(change.getTreeNodeTarget().getUUId());
    			}else
    			if(change.getType() == TreeNodeChangeEventType.CHANGE){
    				TreeNode node = change.getTreeNodeTarget();
    				if(node.isDisposed()){
    					if(!deleted.contains(node))
    						deleted.add(node.getUUId());
    				}else{
    					lastchange.put(node.getUUId(), i);
    					if(!changeParams.containsKey(node.getUUId()))
    						changeParams.put(node.getUUId(), new HashSet<String>());
    					changeParams.get(node.getUUId()).addAll((Set<String>)change.getArguments()[1]);
    				}
    			}else
    			if(change.getType() == TreeNodeChangeEventType.MOVE){
    				TreeNode node = change.getTreeNodeTarget();
    				if(node.isDisposed()){
    					if(!deleted.contains(node))
    						deleted.add(node.getUUId());
    				}//else
    					//lastmove.put(node.getUUId(), i);
    			}else{
    				TreeNode node = change.getTreeNodeTarget();
    				if(node.isDisposed()){
    					if(!deleted.contains(node))
    						deleted.add(node.getUUId());
    				}else
    					created.add(node.getUUId());
    			}
    			i++;
    		}
    		for(UUID uid:deleted){
    			lastchange.remove(uid);
    			//lastmove.remove(uid);
    			created.remove(uid);
    		}
    		for(UUID uid:created){
    			lastchange.remove(uid);
    			//lastmove.remove(uid);
    		}
    		i=0;
        	Set<Integer> events = new HashSet<Integer>();
        	events.addAll(lastchange.values());
        	//events.addAll(lastmove.values());
        	List<TreeNodeChangeEvent> retevents = new LinkedList<TreeNodeChangeEvent>();
        		
    		while(!getGuiChanges().isEmpty())
    		{

    			TreeNodeChangeEvent change = getGuiChanges().poll();
    			switch(change.getType()){
    				case CHANGE:{ 
    					TreeNode node = change.getTreeNodeTarget();
    					try{
	    					if(events.contains(i)){
	    						Set<String>params = changeParams.get(node.getUUId());
	    						Object[]args= change.getArguments();
	    						args[1] = params;
	    						retevents.add(new TreeNodeChangeEvent(TreeNodeChangeEventType.CHANGE, node, args));
	    						change.getTreeNodeTarget().fireGUINodeAttributeChange(params);
	    					}
    					}
    					catch(NullPointerException p){
    						p.printStackTrace();
    					}
    				}break;
    				case CREATE:{
    					TreeNode node = change.getTreeNodeTarget();
    					if(node!=null){
    						retevents.add(change);
    			        }
    					//change.getTreeNodeTarget().fireNodeAttributeChange(
    						//	change.getArguments().length>1?(String)change.getArguments()[1]:TreeNode.ATTR_UNDEFINED);
    				}break;
    				case DELETE:{
    					retevents.add(change);
    					change.getTreeNodeTarget().fireGUINodeDelete();
    				}break;
    				case MOVE:{
    					TreeNode node = change.getTreeNodeTarget();
    				    //if(events.contains(i)){
    				    	retevents.add(change);
    			        //}
    					//change.getTreeNodeTarget().fireNodeAttributeChange(
    						//	change.getArguments().length>2?(String)change.getArguments()[2]:TreeNode.ATTR_UNDEFINED);
    				}break;
    			}
    			i++;
    		}
    		if(retevents.size()>0)
    			for (ITreeChangeListener l : getGUIListeners()) {
    				l.changed(retevents);
    			}
    		
    	}
    }

    /**
     * Get specific object delegated for different checks like check on parent-child relations and element identifiers check.  
     * @return TreeLogic object
     */
    public TreeLogic getTreeLogic(){
        if(tlogic == null)
            tlogic = new TreeLogic();
        return tlogic;
    }
    
    private TreeDB() {
    	latechangelisteners = new HashMap<UUID, INodeChangeListener>();
    }
    public TreeNode createChildNode(UUID pUUId, String id, String type) {
    	return createChildNode(pUUId, id, type, EMPTY_TEMPLATE);
    }
/**
 * Method to create child element for element with pUUId identifier.
 * @param pUUId parent UUID for element
 * @param id identifier of element, needs to comply with TreeLogic.isChildIdValidExc
 * @param type type of element, @see {@link TreeNode.getType}
 * @param templ 
 * @return created Element or null
 */
    public TreeNode createChildNode(UUID pUUId, String id, String type, String templ) {
        UUID uuid = uuidGenerator.nextUUId();
        return createChildNode(uuid, pUUId, id, type, templ, false);
    }

    public TreeNode createChildNode(UUID uuid, UUID pUUId, String id,
            String type,boolean ignoreTemplate) {
    	return createChildNode(uuid, pUUId, id, type,EMPTY_TEMPLATE,ignoreTemplate);
    }
    /**
     * Method to create child element for element with pUUId identifier.
     * @param uuid UUID of element
     * @param pUUId parent UUID for created element
     * @param id identifier of element, needs to comply with TreeLogic.isChildIdValidExc
     * @param type type of element, @see {@link TreeNode.getType}
     * @return created Element or null
     */
    public TreeNode createChildNode(UUID uuid, UUID pUUId, String id,
            String type, String templ,boolean ignoreTemplate) {
        if (getNode(pUUId).findChild(id) != null) {
            throw new IllegalArgumentException("Child with id: \"" + id
                    + "\" exists in " + getNode(pUUId).getQualifiedId());
        }
        storage.createNode(pUUId, uuid, type, id);
        TreeNode node = getNode(uuid), rnode = getRefTarget();
        if(getRefTargetState(rnode)==0)
        	rnode = null;
        TreeNode template = null;
        if(templ == null || !templ.equals("&empty"))
        {
        	template = getTemplate(templ);
        }else
        	template = getActiveTemplate(type);
        boolean changed = false;
        if(node!=null && template!=null && !ignoreTemplate){
        	for(Attribute a:template.getAttributes().values()){
        		if(node.getAttributeKeys().contains(a.getKey()))
        			continue;
        		changed = true;
        		if(a.getOrigin().equals(template.getUUId()))
        			node.putAttribute(a.copy(node.getUUId()));
        		else
        			node.putAttribute(a.copy());
        	}
        }
        if(node!=null&&node.getType().equals(Requirement.TYPE_NAME)&&rnode!=null){
        	changed = true;
        	node.putAttribute(new Attribute(node.getUUId(),AttributeType.REFERENCE,getDefReferenceName(), rnode.getUserFriendlyName()));
        	node.getLinkAttributeValue(getDefReferenceName()).saveAttributes();;
        	if(getRefTargetState(rnode)==1){
        		setRefTarget(null);
        		fireNodeGUIChange(rnode,new HashSet<String>(){{add(getDefReferenceName());}}, true);
        	}
        	node.saveAttributes();
        }
        return node;
    }
    /**
     * Recursively remove element and all its descendants.
     * Child elements are removed first.
     * @param node element to remove
     */
    public void removeNode(TreeNode node) {
    	startChangeRow();
        // Remove children first
    	int i = 0;
        while (node.sizeChildren() > i) {
        	int c = node.sizeChildren();
        	if(storage.canLoad(node.getChild(i).getUUId()))
        		if(node.getChild(i).getType().equals(VirtualNode.TYPE_NAME))
        			node.getChild(i).delete();
        		else
        			removeNode(node.getChild(i));
        	else
        		VirtualNode.getVirtualTreeDB(this).removeNode(node.getChild(i));
        	if(c == node.sizeChildren())
        		i++;
        }
        UUID uuid = node.getUUId();
        //if(first)
        storage.remove(uuid);
        //else
        	//storage.getDeltaTarget().deleted(uuid, true);

        node.setDisposed();
        stopChangeRow();
    }
    /**
     * Move element from one parent to another
     * @param node node to move
     * @param newParent new parent of element
     */
    public void moveNode(TreeNode node, TreeNode newParent) {	
        storage.moveNodeTo(node.getUUId(), newParent.getUUId());
    }
    /**
     * Set identifier of node to id
     * @param node element to set new identifier
     * @param id new identifier
     */
    void setId(TreeNode node, String id) {
        if (!node.getId().equals(id)) {
            storage.changeNodeId(node.getUUId(), id);
            node.fireNodeAttributeChange(new HashSet<String>(){{add("_id");}});
        }
    }

    /**
     * @return root node of TreeDB
     */
    public TreeNode getRootNode() {
    	if (root == null) {
            UUID rootUUID;
            if((rootUUID = storage.getRootNode())==null) { // root doesn't exists  
                return null;
            }
            root = getNode(rootUUID);
        }
        return root;
    }
    
    /**
     * Method to update cached requirement name.
     * @param node requirement to update name
     */
    public synchronized void updateRequirementName(TreeNode node){
        if(!(node instanceof Requirement))
            return;
        Requirement r = (Requirement)node;
        updateRequirementName(node.getUUId(), r.getName());
    }
    /**
     * Method to update cached requirement name.
     * @param node requirement to update name
     */
    public synchronized void updateRequirementName(UUID uid, String newName){
    	Set<UUID> list;
        String oldName = "";
        //Update cache - remove old name
        if(backnamecache.containsKey(uid)){
            oldName = backnamecache.get(uid);
            if(!oldName.equals("") && oldName.equals(newName))
                return;
            list = namecache.get(oldName);
            list.remove(uid);
            namecache.put(oldName, list);
        }
        //if we have new name is not set, stop  
        if(newName == null)
            return;
        //if we have name, add it to cache
        list = namecache.get(newName);
        if(list == null)
            list = new HashSet<UUID>();
        list.add(uid);
        namecache.put(newName, list);
        backnamecache.put(uid, newName);
    }
    /**
     * Method to get all possible elements with given User-Visible name
     * Method to get Array of UUId's of elements with given User Visible Id
     * @param path array of uuids
     */
    public synchronized String[] getNodesByUFName(String path)
    {
    	return getNodesByUFName(path, null);
    }
    public UUID findNode(TreeNode node, String[]tokens ){
    	TreeNode temp = node;
    	for(int i = 0;i<tokens.length&&temp!=null;i++)
    		temp = temp.findChild(tokens[i]);
    	return temp == null?null:temp.getUUId();
    }
    /**
     * Method to get Array of UUId's of elements with given User Visible Id
     * @param path array of uuids
     */
    public synchronized String[] getNodesByUFName(String path, List<String> types){
    	LinkedList<TreeNode> curnodes = new LinkedList<TreeNode>();
        StringTokenizer st = new StringTokenizer(path, "/");
        String nodeId = st.nextToken();
        //if we have node with name nodeId in requirement then use it
        TreeNode rnode = getNode("/Requirements/"+nodeId);
        if(rnode != null)
        	curnodes.add(rnode);
        //if cache of Requiremens names contains nodeId
        //then try to get appropriate node
       	if (namecache.containsKey(nodeId)) {
       		Set<UUID>nodes = namecache.get(nodeId);
       		if(nodes.size() < 1)
       			return null;
       		for(UUID uid : nodes){
       			TreeNode nde = getNode(uid);
       			if(!curnodes.contains(nde))
       				curnodes.add(nde);
       		}
       	}
        if (curnodes.size() == 0) {
            return new String[0];
        }
        
        String[]tokens = new String[st.countTokens()];
        int i = 0;
        while(st.hasMoreTokens())
        	tokens[i++]=st.nextToken();
        
        Iterator<TreeNode> it = curnodes.iterator();
        LinkedList<String> ll = new LinkedList<String>();
        while(it.hasNext()){
        	TreeNode node = it.next();
        	UUID uid = findNode(node, tokens);
        	if(uid!=null)
        		ll.add(uid.toString());
        }
        return ll.toArray(new String[0]);
    }
    private Map<UUID,INodeChangeListener> latechangelisteners;
    @SuppressWarnings("unchecked")
    /**
     * Method to get element by UUID
     * @param uuid element identifier
     * @return element with given uuid
     */
    public synchronized TreeNode getNode(UUID uuid) {
    	TreeNode node = cache.get(uuid); 
        if (node != null) {
            return node;
        }
        if (storage.canLoad(uuid)) {
            //code below was written to deal with the problem of
            //loops in some case and to add change listeners 
            // for linked Nodes. 
            TreeNode n = createFromStorage(uuid, storage.getNodeType(uuid));
            //fix for event loop
            cache.put(n.getUUId(), n);
            //update of name cache for Requirements
            //Using for UFName supporting.
            if(n instanceof Requirement)
                updateRequirementName(n);
            //Get links if they are contained in 
            //current node
            //n.updateGenerators();
            Attribute a = n.getAttribute(TreeNode.ATTR_LINKS);
            if(a != null){
                ArrayList<String>links = (ArrayList<String>) a.getValue();
                for(String link : links){
                    //Link with * - hidden links
                    //we need to deal with them in
                    //some situations.
                    if(link.startsWith("*"))
                        link = link.substring(1);
                    //link contains UUID, so we use it here
                    UUID uid = UUID.fromString(link.substring(0, link.indexOf('_')));
                    INodeChangeListener ncl = n.getChangeListener();
                    if(ncl == null)
                        continue;
                    if(!n.getUUId().equals(uid)){
                        try{
                            getNode(uid).addChangeListener(ncl);
                        }
                        catch(RuntimeException re){
                            latechangelisteners.put(uid, ncl);
                            //Node not found
                        }
                    }else{
                        n.addChangeListener(ncl);
                    }
                }
            }
            //updateGenerators(n);
            if(latechangelisteners.containsKey(uuid)){
                n.addChangeListener(latechangelisteners.get(uuid));
                latechangelisteners.remove(uuid);
            }
            return n;
        } else {
        	if(getStorage() instanceof TreeInstancer){
        		TreeDB db = VirtualNode.getBaseTree(this);
            	if(db == null)
    	        	//TODO ups
            		throw new RuntimeException("node not found: " + uuid);
            	if(db.getStorage().getRootNode().equals(uuid))
            		return db.root!=null?db.getRootNode():db.root;
            	if(!db.getStorage().canLoad(uuid))
            		throw new RuntimeException("node not found: " + uuid);
            	node = db.getNode(uuid);
            	if (node != null)
            		return node;
            	throw new RuntimeException("node not found: " + uuid);
        	}
        	TreeDB db = VirtualNode.getVirtualTreeDB(this);
        	if(!db.equals(this))
        		return db.getNode(uuid);
        	throw new RuntimeException("node not found: " + uuid);
        }
    }
    /**
     * Method to get element by "QualifiedId" or "User-visible name" or "Full name"
     * @param qId path to element
     * @return element if exists, null otherwise, can throw Exception
     */
    public TreeNode getNode(String qId) {
    	return getNode(qId,false);
    }
    /**
     * Method to get element by "QualifiedId" or "User-visible name" or "Full name".
     * If rawpath parameter is true then search ignores hide mechanics.
     * @param qId path to element
     * @param rawPath if true, ignore hide mechanisms
     * @return element if exists, null otherwise, can throw Exception
     */
    public TreeNode getNode(String qId,boolean rawPath) {
        if (qId.equals("/")) {
            return getRootNode();
        }
        qId = qId.replace((char)160, ' ');
        if(qId.startsWith("/")){
        	StringTokenizer st = new StringTokenizer(qId, "/");
        	TreeNode cur = getRootNode();
        	while (st.hasMoreTokens()) {
        		String nodeId = st.nextToken();
        		int ccount = cur.sizeRawChildren();
        		if(!rawPath)
        			ccount = cur.sizeChildren();
        		TreeNode res = null;
        		for(int i = 0; i<ccount && res == null; i++){
        			TreeNode node = null;
        			if(rawPath)
        				node = cur.getRawChild(i);
        			else
        				node = cur.getChild(i);
        			String name = node.getStringAttributeValue(TreeNode.ATTR_NAME);
        			if(name!=null && name.equals(nodeId))
        				res = node;
        			if(node.getId().equals(nodeId))
        				res = node;
        		}
        		cur = res;
        		if (cur == null) {
                	if(isVirtual())
                		return null;
                	return VirtualNode.getVirtualTreeDB(this).getNode(qId,rawPath);
        		}
        	}
        	return cur;
        }else
        {
        	//node by user visible id
        	String[]nodes = getNodesByUFName(qId);
        	if(nodes == null)
        		return null;
        	if(nodes.length!=1){
        		//if(getVirtualState())
        			//return null;
        		if(nodes.length == 0 && !isVirtual())
        			return VirtualNode.getVirtualTreeDB(this).getNode(qId);
        		if(nodes.length == 1)
        			return VirtualNode.getVirtualTreeDB(this).getNode(nodes[0]);
        	}
        	return getNode(UUID.fromString(nodes[0]));
        }
    }
    /**
     * Save attributes of element with uuid as attributes 
     * @param uuid identifier of element
     * @param attributes
     */
    public void saveAttributes(UUID uuid, Set<String> changes, Collection<Attribute> attributes) {
    	//do not need to save if nothing changes
    	if(changes.isEmpty())
    		return;
        storage.saveAttributes(uuid, changes, attributes);
    }
    /**
     * Method to set content of resource with name 'name' on element node  
     * @param node element to set resource for
     * @param name name of resource
     * @param in content of resource
     * @param data @deprecated
     */
    public void setResourceContent(TreeNode node, String name, InputStream in, Object data) {
        storage.setResourceContents(node.getUUId(), name, in, data);
        //update Requirement name 
        if(node.getType().equals(Requirement.TYPE_NAME))
            updateRequirementName(node);
    }
    /**
     * Delete resource on element node with name "name"
     * @param node element from which resource will be deleted
     * @param name name of resource
     */
    public void deleteResource(TreeNode node, String name) {
        storage.deleteResource(node.getUUId(), name);
    }
    /**
     * Method to check existence of resource with name "name" on element node
     * @param node element where resource can be
     * @param name name of resource to check
     * @return true if resource exists, false otherwise
     */
    public boolean isResourceExists(TreeNode node, String name) {
        return storage.isResourceExists(node.getUUId(), name);
    }

    /**
     * Get name of linked to treeDB Storage
     * @return name of storage
     */
    public String getName() {
        return storage.getName();
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Transaction API methods
    //
    // //////////////////////////////////////////////////////////////////////////
    private ArrayList<ITransactionListener> transactionListeners = new ArrayList<ITransactionListener>();
    
    public boolean startTransaction(String name) {
    	return startTransaction(name,false);
    }
    /**
     * Method to start transaction - cancelable set of changes.
     * Data storage will be replaces by Transaction Storage.
     * All changes will be applied on main storage only after commit.
     * @param name name of transaction to show on undo\redo mechanics
     */
    public boolean startTransaction(String name, boolean check) {
        if (storage instanceof TransactionStorage) {
        	if(check)
        		return false;
            throw new IllegalStateException("transaction \"" + getName()
                    + "\"is started");
        }
        startChangeRow();
        TransactionStorage tStorage = new TransactionStorage(storage,
                getTransactionListeners());
        tStorage.setName(name);
        tStorage.setDeltaTarget(storageDeltaListener);
        if(storage instanceof ResourceStorage)
        	while(!((ResourceStorage)storage).lockProject())
				try {
					Thread.currentThread().join(400);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        storage.setDeltaTarget(nullDeltaListener);
        storage = tStorage;
        return true;
    }
    /**
     * Method to end transaction and apply changes on main storage.
     * @return temporaty storage
     */
    public TransactionStorage commit() {
        if (!(storage instanceof TransactionStorage)) {
            stopChangeRow();
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.apply();
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        if(storage instanceof ResourceStorage)
        	((ResourceStorage)storage).unLockProject();
        stopChangeRow();
        storage.setDeltaTarget(storageDeltaListener);
        return tStorage;
    }
    /**
     * Cancel all changes after startTransaction.
     * If transaction is not started will throw Exception
     */
    public void rollback() {
        if (!(storage instanceof TransactionStorage)) {
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.rollback(tStorage);
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        if(storage instanceof ResourceStorage)
        	((ResourceStorage)storage).unLockProject();
        stopChangeRow();
        storage.setDeltaTarget(storageDeltaListener);
//        for (ITransactionListener tl : transactionListeners) {
//            tl.undo(tStorage);
//        }
    }
    /**
     * Add listener on transactions events 
     * @param l listener to add
     */
    public void addTransactionListener(ITransactionListener l) {
    	TreeDB root = VirtualNode.getRootTree(this);
    	if(root.equals(this))
    		transactionListeners.add(l);
    	else
    		root.addTransactionListener(l);
    }
    /**
     * Remove listener from transaction listeners 
     * @param l listener to remove
     */
    public void removeTransactionListener(ITransactionListener l) {
    	TreeDB root = VirtualNode.getRootTree(this);
    	if(root.equals(this))
    		transactionListeners.remove(l);
    	else
    		root.removeTransactionListener(l);
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // Types operations
    //
    // ////////////////////////////////////////////////////////////////////////
    public void addNodeType(String name, Class<? extends TreeNode> clazz) {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			db.addNodeType(name,clazz);
    	}
    	else{
    		NodeType nt = new NodeType(name, TreeDB.this);
    		nt.setClass(clazz);
    		defaultTypes.put(name, nt);
    	}
    }

    public void addNodeType(Class<? extends TreeNode> clazz) {
        if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			db.addNodeType(clazz);
        }else
        	addNodeType(clazz.getSimpleName(), clazz);
    }

    public void addNodeTypes(Map<String, Class<? extends TreeNode>> nts) {
        if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			db.addNodeTypes(nts);
        }else
        	for (String name : nts.keySet()) {
        		addNodeType(name, nts.get(name));
        	}
    }
    /**
     * Method to get map of element types in form
     * of pairs String type name -> NodeType
     * @return available element types
     */
    public Map<String, NodeType> getTypes() {
        if(isVirtual){
        	TreeDB db = VirtualNode.getBaseTree(this);
        	if(db!=null)
        		return db.getTypes();
        }
        Map<String, NodeType> result = new HashMap<String, NodeType>(
                defaultTypes);
        TreeNode typesNode = getNode("/Types",true);
        if (typesNode != null && typesNode.sizeChildren() > 0) {
            int num = typesNode.sizeChildren();
            for (int i = 0; i < num; i++) {
                NodeType node = (NodeType) typesNode.getChild(i);
                result.put(node.getId(), node);
            }
        }
        return result;
    }

    public void createType(NodeType newType) {

        if(isVirtual)
        	VirtualNode.getBaseTree(this).createType(newType);
        else{
        	TreeNode types = getNode("/Types",true);
        	if (types == null) {
        		types = getRootNode().createChild("Types", NodeType.TYPE_NAME, null);
        	}
        	TreeNode newTypeNode = types.createChild(newType.getName());
        	newTypeNode.saveAttributes();
        }
    }

    public NodeType getNodeType(String name) {
        if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.getNodeType(name);
        }
        for (NodeType type : getTypes().values()) {
            if (type.getId().equals(name)) {
                return type;
            }
        }
        return null;
    }

    public boolean removeType(String typeName) {
        if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.removeType(typeName);
        }
       	TreeNode types = getNode("/Types",true);
       	if (types == null) {
       		return false;
       	}
       	TreeNode n = types.findChild(typeName);
       	n.delete();
       	return true;
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // TreeDB factories
    //
    // ////////////////////////////////////////////////////////////////////////

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @param uuidGenerator
     * @return
     */
    public static TreeDB create(String name, ITreeStorage storage,
            IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        storage.createStorage(name);
        storage.setDeltaTarget(newDB.storageDeltaListener);
        newDB.storage = storage;
        // create root node
        UUID uuid = uuidGenerator.nextUUId();
        storage.createNode(null, uuid, "DBRoot", "/");
        newDB.tlogic = new TreeLogic();
        return newDB;
    }

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @return treeDB object
     */
    public static TreeDB create(String name, ITreeStorage storage) {
        return create(name, storage, new DefaultUUIDGenerator());
    }

    public static TreeDB load(ITreeStorage storage, IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        // FIXME: for test we can use mock target, so we left old one if it is
        // not null. Its look like bad design.
        if (storage.getDeltaTarget() == null) {
            storage.setDeltaTarget(newDB.storageDeltaListener);
        }
        newDB.storage = storage;
        return newDB;
    }
    /**
     * Update cache of requirements names.
     * Locate requirements and save its names. 
     */
    public void updateRequirementsNames() {
        DeepFirstTreeWalker tw = new DeepFirstTreeWalker(getRootNode());
        Iterator<TreeNode> it = tw.iterator();
        while(it.hasNext()){
        	TreeNode tn = it.next();
        	if(tn.getType().equals(Requirement.TYPE_NAME))
        		updateRequirementName(tn);
        }
	}
/*
    public void updateRequirementNames(UUID node) {
    	Collection<UUID> children = storage.getChildren(node);
    	for(UUID uid:children){
    		Collection c = storage.getAttributes(uid);
    		String name = null, type = "";
    		Iterator<Attribute> it = c.iterator();
    		while(it.hasNext()){
    			Attribute a = it.next();
    			if(a.getKey().equals(TreeNode.ATTR_TYPE))
    				type = (String)a.getValue();
    			if(a.getKey().equals(TreeNode.ATTR_NAME))
    				name = (String)a.getValue();
    		}
    		if(type.equals(Requirement.TYPE_NAME))
    			updateRequirementName(uid, name);
    		updateRequirementNames(uid);
    	}
	}
*/
	public static TreeDB load(ITreeStorage storage) {
        return load(storage, new DefaultUUIDGenerator());
    }

    // ////////////////////////////////////////////////////////////////////////
	/**
	 * Add listener on element changes
	 * @param listener listener on element changes
	 * @return true if listener is added, false otherwise
	 */
    public boolean addListener(ITreeChangeListener listener) {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.addListener(listener);
    	}
        return listeners.add(listener);
    }
    /**
     * Remove listener on element changes
     * @param listener listener to remove
     * @return true if listener is removed, false otherwise
     */
    public boolean removeListener(ITreeChangeListener listener) {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.removeListener(listener);
    	}
        return listeners.remove(listener);
    }
    /**
     * Get list of listeners on tree changes.
     * @return list of listeners
     */
    public List<ITreeChangeListener> getListeners() {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.getListeners();
    	}
    	return Collections.unmodifiableList(listeners);
    }
    public List<ITreeChangeListener> getGUIListeners() {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.getGUIListeners();
    	}
    	return Collections.unmodifiableList(guilisteners);
    }
    // ////////////////////////////////////////////////////////////////////////
    public boolean addGUIListener(ITreeChangeListener listener) {
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.addGUIListener(listener);
    	}
    	return guilisteners.add(listener); 
    }

    public boolean removeGUIListener(ITreeChangeListener listener) {
    	if(isVirtual)
    		return VirtualNode.getBaseTree(this).removeGUIListener(listener);
    	return guilisteners.remove(listener); 
    }

    private IStorageDeltaListener nullDeltaListener = new IStorageDeltaListener() {
        @Override
        public void moved(UUID uuid, UUID oldParent, UUID newParent,
                boolean internal) {
        }

        @Override
        public void deleted(UUID uuid, boolean internal) {
        }

        @Override
        public void created(UUID uuid, boolean internal) {
        }

		@Override
		public void changed(UUID uuid, Set<String> what, boolean forceInternal) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void resourceChanged(UUID uuid, String resName,
				boolean forceInternal) {
			// TODO Auto-generated method stub
			
		}
    };

    private void notifyExternalChange() {
        for (ITransactionListener tl : transactionListeners) {
            tl._dispose();
        }
    }

    private boolean compareNodes(TreeNode node, NodeDesc nodeDesc) {
        if (!node.getUUId().equals(nodeDesc.getUUId()))
            return false;
        UUID nodeParentUUId = node.getParentUUId();
        UUID descParentUUId = nodeDesc.getParentUUId();
        if (nodeParentUUId == null && descParentUUId != null || nodeParentUUId != null && descParentUUId == null)
            return false;
        if (nodeParentUUId != null && descParentUUId != null && !nodeParentUUId.equals(descParentUUId))
            return false;
        if (!node.getId().equals(nodeDesc.getId()))
            return false;
        
        Map<String, Attribute> oldMap = node.getAttributes();
        Map<String, Attribute> newMap = new HashMap<String, Attribute>(nodeDesc.getAttributes());
        for (String attr_key : oldMap.keySet()) {
            Attribute oldAttr = oldMap.get(attr_key);
            Attribute newAttr = null;
            if (newMap.containsKey(attr_key))
                newAttr = newMap.remove(attr_key);
            else
                return false;
            if (oldAttr == null && newAttr != null || oldAttr != null && newAttr == null)
                return false;
            if (oldAttr != null && newAttr != null && !oldAttr.weakEquals(newAttr))
                return false;
        }
        if (!newMap.isEmpty())
            return false;
        
        List<UUID> oldList = node.getChildrenUUIds();
        List<UUID> newList = new ArrayList<UUID>(nodeDesc.getChildrenUUId());
        for (UUID uuid : oldList) {
            if (newList.contains(uuid))
                newList.remove(uuid);
            else
                return false;
        }
        if (!newList.isEmpty())
            return false;
        return true;
    }

    private IStorageDeltaListener storageDeltaListener = new IStorageDeltaListener() {
        @Override
        public void created(UUID uuid, boolean forceInternal) {
            TreeNode curNode = cache.get(uuid);
            NodeDesc newNodeDesc = loadNodeDescription(uuid);
            if (curNode != null) {
                if (!compareNodes(curNode, newNodeDesc)) {
                    throw new IllegalStateException("node with uuid: "
                            + curNode.getUUId() + "(" + curNode.getUserFriendlyName()
                            + ") exists");
                } else {
                    return;
                }
            }
        	startChangeRow();
            curNode = createModelNode(
                    newNodeDesc.getUUId(), 
                    newNodeDesc.getParentUUId(), 
                    newNodeDesc.getId(), 
                    (String)newNodeDesc.getAttributes().get(TreeNode.ATTR_TYPE).getValue(),
                    newNodeDesc.getAttributes().values(),
                    newNodeDesc.getChildrenUUId());
            
            TreeNode parent = curNode.getParent(true);
            if (parent != null && !parent.containsChildUUID(uuid))
                parent.addChildUUID(uuid); 
            cache.put(uuid, curNode);
            if (curNode.getId().equals("/")) {
                assert curNode.parent == null;
                root = curNode;
            }
            //updateGenerators(curNode);
            if (!forceInternal) 
                notifyExternalChange();
            addGuiChange(new TreeNodeChangeEvent(TreeNodeChangeEventType.CREATE, curNode, forceInternal));
            for (ITreeChangeListener l : getListeners()) {
	            l.created(curNode, forceInternal);
    	    }
            
            //update Requirement name in cache
            if(curNode instanceof Requirement)
                updateRequirementName(curNode);
            stopChangeRow();
        }

		private NodeDesc loadNodeDescription(UUID uuid) {
            String typeName = storage.getNodeType(uuid);
            UUID parentUUId;
            String id;
            if (typeName.equals("DBRoot")) {
                id = "/";
                parentUUId = null;
            } else {
                id = storage.getNodeId(uuid);
                parentUUId = storage.getParent(uuid);
            }
            Map<String, Attribute> attributes = new TreeMap<String, Attribute>();
            Collection<Attribute> attrs = storage.getAttributes(uuid);
            for (Attribute a : attrs) {
                attributes.put(a.getKey(), a);
            }
            List<UUID> childrenUUId = new ArrayList<UUID>(
                    storage.getChildren(uuid));
            return new NodeDesc(uuid, parentUUId, id, attributes, childrenUUId);
        }
        /**
         * Reaction on element deletion
         */
        @Override
        public void deleted(UUID uuid, boolean internal) {
            TreeNode node = cache.get(uuid);
            //if node is not cache
            if(node == null){
                return;
            }
        	startChangeRow();
            NodeDesc nodeDesc = NodeDesc.fromNode(node);
            try{
            	TreeNode par = node.getParent(true);
            	if(par!=null)
            		par.removeChildUUID(uuid);
            }
            catch(RuntimeException re){}
            //refresh cache
            if (backnamecache.containsKey(uuid)){
                node.removeAttribute(Requirement.ATTR_NAME);
                updateRequirementName(node);
            }
        	addGuiChange(new TreeNodeChangeEvent(TreeNodeChangeEventType.DELETE, node, internal, nodeDesc));
            node.fireNodeDelete();
            cache.remove(uuid);
            node.setDisposed();
            if (!internal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : getListeners()) {
                l.deleted(nodeDesc, internal);
            }

            stopChangeRow();
        }

        /**
         * Reaction on element change
         */
        @Override
        public void changed(UUID uuid, Set<String> what, boolean internal) {
            // Sometimes (e.g, import of existing project)
        	// changed event happens when 
        	// there is no node in the cache yet.
        	// So, use getNode(uuid) instead cache.get(uuid).
        	startChangeRow();
        	TreeNode curNode = getNode(uuid);
        	if(what == null)
        		what = curNode.getAttributeKeys();
            String newId = storage.getNodeId(uuid);
            boolean neq = !newId.equals(curNode.id);
            curNode.id = newId;
        	if (curNode instanceof Requirement && neq) {
        	    ((Requirement)curNode).updateTreeNodes((Requirement)curNode, "_id");
        	}
        	//todo use what
            loadNodeAttributes(uuid, curNode, what);
            //for(String a : curNode.getAttributeKeys()){
        	//}
       		VirtualNode.updateClonesAttributes(curNode, curNode.getAttributes(), false);
            addGuiChange(new TreeNodeChangeEvent(TreeNodeChangeEventType.CHANGE, curNode, internal, what));
            if (!internal) {
                notifyExternalChange();
            }
            fireNodeChange(curNode,what, internal);
            fireNodeGUIChange(curNode,what, internal);
            if(curNode != null){
            	curNode.fireNodeAttributeChange(what);
            	//TODO remove?
            	curNode.fireGUINodeAttributeChange(what);
            }
            stopChangeRow();
        }

        @Override
        public void moved(UUID nodeUUID, UUID oldParentUUId, UUID newParentUUId,
                boolean internal) {
        	startChangeRow();
        	TreeNode node = getNode(nodeUUID);
            TreeNode newParent = getNode(newParentUUId);
            TreeNode oldParent = getNode(oldParentUUId);

            node.parent = newParentUUId;
            if(oldParent!=null)
                oldParent.removeChildUUID(nodeUUID);
            newParent.addChildUUID(nodeUUID);

            if (!internal) {
                notifyExternalChange();
            }
            addGuiChange(new TreeNodeChangeEvent(TreeNodeChangeEventType.MOVE, node, internal, oldParent == null ? node.getParent(true) : oldParent));
            for (ITreeChangeListener l : listeners) {
                l.moved(node, oldParent == null ? node.getParent(true) : oldParent,
                        internal);
            }
            stopChangeRow();
        }

		@Override
		public void resourceChanged(UUID uuid, String resName,
				boolean forceInternal) {
			// TODO Auto-generated method stub
			
		}
    };

    private void loadNodeAttributes(UUID uuid, TreeNode node, Set<String> what) {
        Collection<Attribute> newAttributes = storage.getAttributes(uuid);
        applyNodeAttributes(node, newAttributes, what);
    }

    @SuppressWarnings("unchecked")
    private void applyNodeAttributes(TreeNode node, Collection<Attribute> newAttributes, Set<String> what) {
        //for new and old Locations of Requirement reset labels
        if (node instanceof Requirement) {
            Requirement req = (Requirement) node;
            String oldName = req.getName();
            String newName = null;
            //get old Locations list
            List<String> oldLocs = new ArrayList<String>(req.getLocationQIds());
            //construct new Locations list
            List<String> newLocs = null;
            boolean locremoved = false;
            if(what.contains(Requirement.ATTR_LOCATIONS) || what.contains(TreeNode.ATTR_NAME)){
	            for (Attribute attr : newAttributes) {
	                if (attr.getKey().equals(Requirement.ATTR_LOCATIONS)) {
	                    Object val = attr.getValue();
	                    if (val instanceof List) {
	                        newLocs = new ArrayList<String>((List<String>) val);
	                    }
	                } else
	                if (attr.getKey().equals("_name")) {
	                    newName = (String) attr.getValue();
	                }
	            }
	            if(what.contains(Requirement.ATTR_LOCATIONS)&&newLocs == null)
	            	locremoved = true;
	            	
            }
            //compare lists
            boolean fNewLoc = false;
            if (newLocs != null) {
                for (String newLoc : newLocs) {
                    if (!oldLocs.remove(newLoc)) {
                        fNewLoc = true;
                    }
                }
            }
            //and call affected Locations
            if (fNewLoc || !oldLocs.isEmpty()) {
                //remaining old locations (removed)
                for (String oldLoc : oldLocs) {
                    TreeNode locNode = getNode(oldLoc);
                    if (locNode instanceof Location) {
                        ((Location)locNode).resetLabel();
                    }
                }
                //all other locations
                if (newLocs != null) {
                    for (String newLoc : newLocs) {
                        TreeNode locNode = getNode(newLoc);
                        if (locNode instanceof Location) {
                            ((Location)locNode).resetLabel();
                        }
                    }
                }
            }
            //check, if name was changed
        }
        Map<String, List<Attribute>> oldAttributes = node.attributes;
        if(oldAttributes!=null)
        	for(String s:what)
        		if(oldAttributes.containsKey(s))
        		{
        			List<Attribute> attrs = oldAttributes.get(s);
        			if(!attrs.isEmpty())
        				node.reactBeforeAttributeChange(s, attrs.get(0));
        		}
        updateNodeAttributes(node, newAttributes,what);
        Set<String> keys = new HashSet<String>(oldAttributes.keySet());
        for(Attribute attribute : newAttributes){
        	if(what.contains(attribute.getKey())){
        		Attribute old = null;
        		if(oldAttributes.containsKey(attribute.getKey())&&oldAttributes.get(attribute.getKey()).size()>0)
        			old = oldAttributes.get(attribute.getKey()).get(0);
        		node.reactAfterAttributeChange(attribute.getKey(),old, attribute);
        	}
        	keys.remove(attribute.getKey());
        }
        for(String s:keys)
        	if(what.contains(s)){
        		Attribute old = null;
        		if(oldAttributes.containsKey(s)&&oldAttributes.get(s).size()>0)
        			old = oldAttributes.get(s).get(0);
        		node.reactAfterAttributeChange(s,old, null);
        	}
    }
    /**
     * Method to replace node attributes by stored values
     * @param node element to replace attributes on
     */
    public void updateNodeAttributes(TreeNode node){
    	Set<String> keys = new HashSet<String> (node.getAttributeKeys());
    	Collection<Attribute> stA = storage.getAttributes(node.getUUId());
    	for(Attribute a:stA)
    		keys.add(a.getKey());
    	updateNodeAttributes(node, stA,keys);
    }
    
    private void updateNodeAttributes(TreeNode node, Collection<Attribute> newAttributes, Set<String> what) {
        HashSet<String> removeKeys = new HashSet<String>(
                node.attributes.keySet());
        UUID uid = node.getParentUUId(true);
        boolean isRoot = (uid==null || uid.equals(root));
        TreeNode basenode = node.getCloneParent();
        if(basenode!=null)
        	for(String s:what)
        		if(basenode.getAttributes().containsKey(s))
        			for(Attribute a:basenode.attributes.get(s))
        				node.putAttribute(a,true,false);
        
        if(!isRoot){
        	TreeNode par = node.getParent(true);
        	for(String s:what)
        		if(par.getAttributes().containsKey(s))
        			for(Attribute a:par.attributes.get(s))
        				if(!a.getKey().startsWith("__") && a.getAvailability()!=ParameterAvailability.LOCAL)
        					node.putAttribute(a,true,false);
        }
        
        Set<TreeNode> settoup = new HashSet<TreeNode>();
        for (Attribute a : newAttributes) {
        	if(!what.contains(a.getKey()))
        		continue;
        	if(node.attributes.containsKey(a.getKey())&&node.attributes.get(a.getKey()).size()>0
        			&&node.attributes.get(a.getKey()).get(0).isMoreVisible(a, node.getUUId()))
        		node.removeAttribute(node.attributes.get(a.getKey()).get(0).getOrigin(),a.getKey(), false, true);
            removeKeys.remove(a.getKey());
            if(isRoot || a.getKey().startsWith("__"))
            	node.rawPutAttribute(a);
            else{
            	node.putAttribute(a,true,false);
            	settoup.addAll(node.propogateAttributeChange(a));
            }
        }
        for(TreeNode nd:settoup)
        	nd.saveAttributes();
        for (String key : removeKeys) {
        	if(!what.contains(key))
        		continue;
        	node.removeAttribute(node.getUUId(), key,true,false);
        }
        if(node instanceof Requirement)
            updateRequirementName((Requirement)node);
    }
    /**
     * Method to get uuid generator 
     * @return uuid generator
     */
    public IUUIDGenerator getUuidGenerator() {
        return uuidGenerator;
    }
    /**
     * Get current data storage. It may be related to one of possible storage mechanisms.
     * Most frequently used are Resource storage based on eclipse mechanisms and
     * Transaction storage - temporary storage during transactions. One more variant is
     * TreeInstancer that store data during report generation 
     * @return current data storage
     */
    public ITreeStorage getStorage() {
        return storage;
    }
    private boolean defaultTypesContainsKey(String typeName){
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.defaultTypesContainsKey(typeName);
    	}
    	return defaultTypes.containsKey(typeName);
    }
    private NodeType defaultTypesGet(String typeName){
    	if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			return db.defaultTypesGet(typeName);
    	}
    	return defaultTypes.get(typeName);
    }
    private TreeNode createModelNode(UUID uuid, UUID puuid, String id,
            String typeName, Collection<Attribute> attributes, Collection<UUID> children) {
        TreeNode result = null;
        if (defaultTypesContainsKey(typeName)) {
            NodeType nt = defaultTypesGet(typeName);
            try {
                Class<?>[] treeNodeArgs = new Class[] { TreeDB.class,
                        UUID.class };
                result = nt.getTypeClass().getConstructor(treeNodeArgs)
                        .newInstance(TreeDB.this, uuid);
                result.setType(typeName);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (result == null) {
            result = new TreeNode(this, uuid);
            result.setType(typeName);
        }
        result.id = id;
        result.parent = puuid;
        //TODO check
        cache.put(uuid, result);
        Set<String> keys = new HashSet<String>();
        for(Attribute a:attributes)
        	keys.add(a.getKey());
        updateNodeAttributes(result, attributes, keys);
        updateNodeChildren(result,children, true);
        return result;
    }

	public void updateNodeChildren(TreeNode result) {
		updateNodeChildren(result,false);
    }
    
	public void updateNodeChildren(TreeNode result, boolean init) {
		if(result==null || !storage.canLoad(result.getUUId()) || storage.getChildren(result.getUUId()) == null)
			return;
		updateNodeChildren(getNode(result.getUUId()), storage.getChildren(result.getUUId()),init);
    }
    
    private void updateNodeChildren(TreeNode result, Collection<UUID> children, boolean init) {
    	List<UUID> ids = new ArrayList<UUID>(children);
        result.setChildrenUUIDs(ids);
        if(!init){
        	//change update?
        	fireNodeChange(result,new HashSet<String>(){{add("_id");}}, true);
        }
        //result.getChildrenUUIds();
//        getDeltaTarget().changed(result.getUUId(), true);
	}

	private TreeNode createFromStorage(UUID uuid, String typeName) {
        String id;
        UUID puuid;
        if (typeName.equals("DBRoot")) {
            id = "/";
            puuid = null;
        } else {
            id = storage.getNodeId(uuid);
            puuid = storage.getParent(uuid);
        }
        Collection<Attribute> attributes = storage.getAttributes(uuid);
        Collection<UUID> children = storage.getChildren(uuid);
        if(id == null)
        	return null;
        TreeNode result = createModelNode(uuid, puuid, id, typeName, attributes, children);
        if (result.getId().equals("/")) {
            assert result.parent == null;
            root = result;
        }
        
        return result;
    }
    /**
     * Method to fire element change event. It will be translated to all subscribed listeners.
     * @param curNode changed element
     * @param internal true if event appears from inner mechanics
     */
    public void fireNodeChange(TreeNode curNode, Set<String> params, boolean internal) {
    	if(isVirtual && curNode!=null){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			db.fireNodeChange(curNode, params, internal);
    	}
    	else
    		for (ITreeChangeListener l : listeners) {
    			l.changed(curNode, params, internal);
    		}
    }
    private static List<String> typeFilter = new ArrayList<String>();
    /**
     * String for representing of Node's types filter in form of list of string names of types. For ex. {TestPurpose.TYPE_NAME}"
     * @return
     */
    public List<String> getLinkToTypeFilter(String linktype){
        //TODO TreeLogic methods
        if(linktype.equals("TestStepLinks")){
            ArrayList<String> a = new ArrayList<String>();
            a.add(TestPurpose.TYPE_NAME);
            return a;
        }
        return typeFilter;
    }
    public List<String> getRelatedLinkTo(){
        ArrayList<String> a = new ArrayList<String>();
        a.add("TestStepLinks");
        return a;
    }
    public List<String> getLinkToTypeFilter(){
        //TODO TreeLogic methods
        List<String>types = new ArrayList<String>() ,
                lt = getRelatedLinkTo();
        for(String s : lt)
            types.addAll(getLinkToTypeFilter(s));
        return types;
    }
    static final Pattern new_linkto_pattern = Pattern.compile("\\[linkto: ([^\\]]*)\\]", Pattern.MULTILINE & Pattern.DOTALL);
    static final Pattern linkto_pattern = Pattern.compile("\\[<span id=\"([a-zA-Z0-9-]+)\">.{0,10}</span>linkto: ([^\\]]*)\\]", Pattern.MULTILINE & Pattern.DOTALL);
    /**
     * Update stored linkto: constructions and attribute where
     * [linkto: ] data stored.
     * @param attrname Name of attribute where links can be placed 
     */
    @SuppressWarnings("unchecked")
    public void updateLinks(TreeNode node, String attrname, boolean removeHiddenLinks){
//        boolean changed = false;
//        String newValue = node.getStringAttributeValue(attrname),oldValue = newValue+"";
//        if(newValue == null)
//            return;
//        //get change listener of this node
//        INodeChangeListener changeListener = node.getChangeListener();
//        TreeDB db = node.getTreeDB();
//        //get links of this node
//        Attribute a = node.getAttribute(TreeNode.ATTR_LINKS);
//        //list of links in format UUID_UFName
//        List<String> attr = null;
//        TreeMap<String,String> set = null;
//        if(a != null)
//            attr = (List<String>)(a.getValue());
//        //find stored links and 
//        Matcher m = linkto_pattern.matcher(newValue);
//        if(m.find()){
//            set = new TreeMap<String,String>();
//            String ufName = "", uid=""; 
//            while(!m.hitEnd()){
//                TreeNode n = null;
//                try
//                {
//                   if(m.group(1) != null){
//                       uid = m.group(1);
//                       //if UUID starts with * then
//                       //element is hidden (link in title)
//                       if(uid.startsWith("*"))
//                           uid = uid.substring(1);
//                       n = db.getNode(UUID.fromString(uid));
//                       //remove old listeners
//                       //n.removeChangeListener(changeListener);
//                   }
//                }
//                catch(Exception e)
//                {
//                }
//                   ufName = m.group(2);
//                   //check if all right
//                   // set some properties to non-default values
//                   if(n != null){
//                       if(ufName.equals(n.getUserFriendlyName()))
//                           set.put(uid, uid + "_" + n.getUserFriendlyName());
//                   }
//                m.find();
//            }
//        }
//        //look for new links and add it
//        Matcher newm = new_linkto_pattern.matcher(newValue);
//        if(newm.find()){
//            //if we have no links before then initiate
//            //set for results
//            if(set == null)
//                set = new TreeMap<String,String>();
//            String ufname = "", uid="";
//            while(!newm.hitEnd()){
//                TreeNode n = null;
//                try
//                {
//                    ufname = newm.group(1);
//                    String[] ar = db.getNodesByUFName(ufname, getLinkToTypeFilter());
//                    //if we have exacly 1 node with given ufname
//                    //then this link is correct
//                    if(ar.length == 1)
//                        n = db.getNode(UUID.fromString(ar[0]));
//                    //n.removeChangeListener(changeListener);
//                }
//                catch(Exception e)
//                {
//                }
//                //if we found one node then add it in
//                //attribute in correct format 
//                if(n != null){
//                    uid = n.getUUId().toString();
//                    set.put(uid, uid + "_" + ufname);
//                    newValue = newValue.replace(newm.group(0), 
//                            "[<span id=\"" + uid + "\">&zwnj;</span>linkto: " + ufname + "]");
//                }
//                newm.find();
//            }
//        }
//        //another option - nodes that are stored in attributes
//        //we need to remove listeners
//        if(set == null)
//           set = new TreeMap<String, String>();
//        if(attr != null && attr.size() > 0){
//            int num = attr.size();
//            for(int i = 0; i < num; i++)
//            {
//                String attr_str = attr.get(i);
//                String temp = attr_str.substring(0, attr_str.indexOf('_'));
//                TreeNode n = null;
//                //we need to save top link in this case
//                //now all links that are not in attribute
//                //will be deleted, but only exception is
//                //"hidden" link related to title link
//                if(temp.startsWith("*"))
//                {
//                    //if we dont want to remove hidden
//                    //link then save it
//                    if(!removeHiddenLinks)
//                        set.put(temp, attr_str);
//                    temp = temp.substring(1);
//                }
//                try
//                {
//                    //try to get node and remove old listener
//                    //from it
//                    n = db.getNode(UUID.fromString(temp));
//                    n.removeChangeListener(changeListener);
//                }
//                catch(Exception e)
//                {
//                }
//            }
//        }
//        //now it is time to save collected
//        //links and add listeners
//        int cnt = 0;
//        for(String s : set.keySet()){
//            try
//            {
//                if(attr != null && attr.contains(set.get(s)))
//                        cnt++;
//                TreeNode n;
//                //"hidden" links are sometimes tricky one
//                if(s.startsWith("*"))
//                    s = s.substring(1);
//                n = db.getNode(UUID.fromString(s));
//                //add listener to node
//                n.addChangeListener(changeListener);
//            }
//            catch(Exception e)
//            {
//            }
//        }
//        if(cnt != (attr==null?0:attr.size()) || cnt != (set==null?0:set.size()))
//            changed = true;
//        //in the case when text changed then we
//        //need to update attribute
//        if(!oldValue.equals(newValue)){
//            node.putAttribute(new Attribute(node, AttributeType.STRING,attrname,newValue));
//            changed = true;
//        }
//        //if we have no link then remove linkto attribute
//        //else replace\create it
//        if (set == null || set.size() == 0){
//            changed = changed || (attr != null && attr.size() != 0);
//            node.removeAttribute(TreeNode.ATTR_LINKS);
//        }else{
//            node.putAttribute(
//                new Attribute(
//                		node, 
//                        AttributeType.LIST,
//                        TreeNode.ATTR_LINKS, new ListValue(AttributeType.STRING, new ArrayList<Object>(set.values()))
//                    )
//                );
//        }
//        if(changed){
//            node.saveAttributes();
//            //fire event "links changed"
//            fireNodeChange(node, true);
//            node.fireNodeAttributeChange(TreeNode.ATTR_LINKS);
//        }
//	disable test step -> req link
    }
    public void updateLinks(TreeNode node, String attrname){
        updateLinks(node, attrname, false);
    }

	public void fireNodeGUIChange(TreeNode curNode,Set<String> attrs, boolean internal) {
		if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null)
    			db.fireNodeGUIChange(curNode, attrs, internal);
		}else
    		if(activeChangesRowsCount == 0)
    			for (ITreeChangeListener l : getGUIListeners()) {
    				l.changed(curNode, attrs, internal);
    			}
    		else
    			addGuiChange(new TreeNodeChangeEvent(TreeNodeChangeEventType.CHANGE, curNode, internal, attrs));
	}

	private void addGuiChange(TreeNodeChangeEvent treeNodeChangeEvent) {
		if(isVirtual){
    		TreeDB db = VirtualNode.getBaseTree(this);
    		if(db!=null){
    			db.addGuiChange(treeNodeChangeEvent);
    			return;
    		}
		}
		guichanges.add(treeNodeChangeEvent);
	}


	public IStorageDeltaListener getDeltaTarget() {
		return storageDeltaListener;
	}

	private LinkResolver lr = null;

	public LinkResolver getLinkResolver() {
		if(lr == null)
			if(isVirtual){
	    		TreeDB db = VirtualNode.getBaseTree(this);
	    		if(db!=null)
	    			return db.getLinkResolver();
			}else
				lr = new LinkResolver();
		return lr;
	}

	public boolean isLoaded(UUID uid) {
		return cache.containsKey(uid)||(!isVirtual()?VirtualNode.getVirtualTreeDB(this).isLoaded(uid):false);
	}

	private TreeNode refTarget;
	private int refTargetState = 0;
	public TreeNode getRefTarget(){
		return refTarget;
	}
	public int getRefTargetState(TreeNode node) {
		if(refTarget != null && refTarget.equals(node))
			return refTargetState;
		return 0;
	}
	public void setRefTarget(TreeNode node){
		if(refTarget == null)
			refTargetState = node!=null?1:0;
		else if(!refTarget.equals(node)){
			refTargetState = 0;
			refTarget = null;
			return;
		}
		if(refTarget != null && refTarget.equals(node))
			refTargetState=(this.refTargetState+1)%3;
		refTarget = node;
		getRootNode().putAttribute(new Attribute(getRootNode(), AttributeType.STRING, "_refTarget",node==null?"":node.getQualifiedId()));
		getRootNode().saveAttributes();
	}
	public String getDefReferenceName(){
		Attribute a = root.getAttribute(DEFAULT_REFERENCE_NAME);
		if(a!=null)
			return a.getValue().toString();
		return "elaborates";
	}

	public int getGlobalRefTargetState() {
		return refTargetState;
	}
	public List<ITransactionListener> getTransactionListeners() {
    	TreeDB root = VirtualNode.getRootTree(this);
    	if(root.equals(this))
    		return transactionListeners;
    	else
    		return root.getTransactionListeners();
	}
	public String getQualifiedId(UUID origin) {
		if(origin==null || !getStorage().canLoad(origin))
			return "";
		String id = getStorage().getNodeId(origin);
		if(id.equals("/"))
			return "";
		return getQualifiedId(getStorage().getParent(origin))+"/"+id;
	}
	public void fireNodeGUIRemoved(TreeNode node, boolean b) {
		node.fireGUINodeDelete();
		for(ITreeChangeListener inc : getGUIListeners())
			inc.deleted(node.toNodeDesc(),true);
	}

	public enum SortingMethod{BY_NAMES,BY_LOCATIONS,BY_INDEX};
    private boolean getSortByLocationsState(){
		if(getRootNode() == null)
			return true;
		Attribute a = getRootNode().getAttribute(TreeNode.SORT_BY_LOCATIONS);
		return a == null || a.getValue().equals("true") || a.getValue().equals(true);
	}
    private boolean getSortSeparatorState(){
		if(getRootNode() == null)
			return true;
		Attribute a = getRootNode().getAttribute(TreeNode.SORT_SEP_TYPES);
		return a == null || a.getValue().equals("true") || a.getValue().equals(true);
	}
    private boolean getSortByIndexState(){
		if(getRootNode() == null)
			return true;
		Attribute a = getRootNode().getAttribute(TreeNode.SORT_BY_INDEX);
		return a != null && (a.getValue().equals("true") || a.getValue().equals(true));
	}
	public SortingMethod getSortingMehod() {
		if(getSortByLocationsState())return SortingMethod.BY_LOCATIONS;
		if(getSortByIndexState()) return SortingMethod.BY_INDEX;
		return SortingMethod.BY_NAMES;
	}
	public String getSortingString() {
		return getSortByLocationsState()+"_"+getSortByIndexState()+"_"+getSortSeparatorState();
	}
	public String getFullName(String targetQID) {
		try{
			return getFullName(getNode(targetQID));
		}catch(Exception e){
			
		}
		return null;
	}

	public String getFullName(TreeNode node) {
		if(node == null)
			return null;
		if(getFullNameMode())
			return node.getUserFriendlyName();
		return  Requirement.getFullName(node, this);
	}
}
